## three dimensional simulation of the three-body problem
## using the Python Visual module for the animation
## see http://www.vpython.org/
from visual import *

global threebody
scene = display(title='University of Glasgow A1 Dynamical Astronomy orbital kick demo (press 1, arrows). RH mouse rotates, LH+RH zooms.', width=1280, height=1080)
scene.autoscale=false;

## set up the masses. m1 is missing in 2-body systems
m1 = sphere(color=color.cyan)
m2 = sphere(color=color.yellow)
m3 = sphere(color=color.white, make_trail=True, interval=2, retain=1000)
pointer = arrow(color=color.orange, pos=vector( 1000,0,0),axis=vector( 1,1,1), shaftwidth=1)

## move to the centre of mass frame and adjust sizes for constant density
def resetscene(m1,m2,m3):
    vcentre=(m1.mass*m1.vel+m2.mass*m2.vel+m3.mass*m3.vel)/(m1.mass+m2.mass+m3.mass)
    for a in [m1, m2, m3]:
        a.vel -= vcentre
        a.radius = 0.5+ 0.05*a.mass**(1.0/3.0)
m3.trail_object.pos=[]

def dydt(y):
## determine the derivative of the state vector y
    deriv = zeros((6,3), dtype=vector)
    r12=y[0]-y[2]; r23=y[2]-y[4]; r31=y[4]-y[0]
    r12c=r12/mag(r12)**3; r23c=r23/mag(r23)**3; r31c=r31/mag(r31)**3
    deriv[1] = -m2.mass*r12c + m3.mass*r31c
    deriv[3] = -m3.mass*r23c + m1.mass*r12c
    deriv[5] = -m1.mass*r31c + m2.mass*r23c
    deriv[0:5:2] = y[1:6:2] ## just copy the three velocities from y
    return deriv


def keyInput(evt):
    global threebody,dt
    s = evt.key
    if (s == '1'): ##circular
        m1.pos=vector( 1000,0,0); m1.vel=vector(0,0,0);   m1.mass=0
        m2.pos=vector( 0,0,0);   m2.vel=vector( 0, 0,0); m2.mass=1000
        m3.pos=vector( 10,0,0);  m3.vel=vector(0,10,0);  m3.mass=0.1
        scene.range=40
        resetscene(m1,m2,m3)
        pointer.pos=vector( 1000,0,0)
        m3.trail_object.pos=[]

        threebody = False;
    if (s == 'up'):
        delta_v = 2*m3.vel/mag(m3.vel)
        m3.vel = m3.vel + delta_v
        pointer.pos=m3.pos
        pointer.axis=delta_v*3
    if (s == 'down'):
        delta_v = -2*m3.vel/mag(m3.vel)
        m3.vel = m3.vel + delta_v
        pointer.pos=m3.pos
        pointer.pos=m3.pos
        pointer.axis=delta_v*3
    if (s == 'left'):
        delta_v = -6*m3.pos/mag(m3.pos)
        m3.vel = m3.vel + delta_v
        pointer.pos=m3.pos
        pointer.axis=delta_v
    if (s == 'right'):
        delta_v = 6*m3.pos/mag(m3.pos)
        m3.vel = m3.vel + delta_v
        pointer.pos=m3.pos
        pointer.axis=delta_v
 
 
scene.bind('keydown', keyInput)

## initiate with the simplest system
m1.pos=vector( 1000,0,0); m1.vel=vector(0,0,0);   m1.mass=0
m2.pos=vector( 0,0,0);   m2.vel=vector( 0, 0,0); m2.mass=1000
m3.pos=vector( 10,0,0);  m3.vel=vector(0,10,0);  m3.mass=0.1
scene.range=40
resetscene(m1,m2,m3)
threebody = False;

dt = 0.02 ## the time step.  Keep this as small as possible.

while True:
  rate(150)
  ## solve using forth-order Runge Kutta.  See CHOPF p62.
  y = [m1.pos,m1.vel,m2.pos,m2.vel,m3.pos,m3.vel]
  k1 = dt*dydt(y);  k2 = dt*dydt(y+k1/2.0)
  k3 = dt*dydt(y+k2/2.0);  k4 = dt*dydt(y+k3)
  dy = k1/6.0 + k2/3.0 +k3/3.0 + k4/6.0
## update the animation
  if threebody:
    m1.pos += dy[0]; m1.vel += dy[1]
  if mag(m2.pos-m3.pos)>1:
     m2.pos += dy[2]; m2.vel += dy[3]
     m3.pos += dy[4]; m3.vel += dy[5]

